-- YGREC8/YGREC8_asm.vhdl
-- version jeu. nov. 16 04:28:58 CET 2017 by Yann Guidon (whygee@f-cpu.org)
-- version ven. nov. 17 01:59:15 CET 2017
-- version sam. nov. 18 06:08:39 CET 2017 : a new approach for the assembler, less painful
-- version dim. nov. 19 07:36:31 CET 2017 : assembkler working
-- Released under the GNU AGPLv3 license
--
-- A collection of functions to assemble binary code for the YGREC8

Library ieee;
    use ieee.std_logic_1164.all;
    use ieee.numeric_std.all;
Library work;
    use work.all;
    use work.ygrec8_def.all;
    use work.ygrec8_utils.all;

package ygrec8_asm is

  -- 6 individual formats for manual assembly:
  function Assemble_RIO  (Op: SLV4; dst: SLV3; port_nr: integer        ) return SLV16;
  function Assemble_RI8  (Op: SLV4; dst: SLV3; val: integer            ) return SLV16;
  function Assemble_RI3  (Op: SLV4; dst: SLV3; val: integer            ) return SLV16;
  function Assemble_RI3C (Op: SLV4; dst: SLV3; val: integer; cond: SLV4) return SLV16;
  function Assemble_RR   (Op: SLV4; dst: SLV3; src: SLV3               ) return SLV16;
  function Assemble_RRC  (Op: SLV4; dst: SLV3; src: SLV3;    cond: SLV4) return SLV16;

  -- this one does it all:
  function Assemble(txt, context: string) return SLV16;

  -- used both for validating alphanumerical characters
  -- and providing an uppercase function
  type char_array_type is array (character) of character;
  constant upcase_lookup : char_array_type := (
   '0'=>'0',  '1'=>'1',  '2'=>'2',  '3'=>'3',  '4'=>'4',
   '5'=>'5',  '6'=>'6',  '7'=>'7',  '8'=>'8',  '9'=>'9',  '-'=>'-',
   'a'=>'A',  'b'=>'B',  'c'=>'C',  'd'=>'D',  'e'=>'E',  'f'=>'F',  'g'=>'G',
   'h'=>'H',  'i'=>'I',  'j'=>'J',  'k'=>'K',  'l'=>'L',  'm'=>'M',  'n'=>'N',
   'o'=>'O',  'p'=>'P',  'q'=>'Q',  'r'=>'R',  's'=>'S',  't'=>'T',  'u'=>'U',
   'v'=>'V',  'w'=>'W',  'x'=>'X',  'y'=>'Y',  'z'=>'Z',
   'A'=>'A',  'B'=>'B',  'C'=>'C',  'D'=>'D',  'E'=>'E',  'F'=>'F',  'G'=>'G',
   'H'=>'H',  'I'=>'I',  'J'=>'J',  'K'=>'K',  'L'=>'L',  'M'=>'M',  'N'=>'N',
   'O'=>'O',  'P'=>'P',  'Q'=>'Q',  'R'=>'R',  'S'=>'S',  'T'=>'T',  'U'=>'U',
   'V'=>'V',  'W'=>'W',  'X'=>'X',  'Y'=>'Y',  'Z'=>'Z',      others=>NUL);

end ygrec8_asm;

package body ygrec8_asm is

  ------------------------------------------------------------------------
  -- some warmup:

  function Assemble_RIO (Op: SLV4; dst: SLV3; port_nr: integer) return SLV16 is
    variable nr : integer := port_nr;
  begin
    assert (Op = Op_IN) or (Op = Op_OUT)
        report "Assemble_IO : wrong opcode :" & SLV_to_bin(Op)
         severity failure;

    assert (nr < 512) and (nr >= -256)
        report "Assemble_IO : port number is out of range :" & integer'image(nr)
         severity failure;
    if nr < 0 then    -- sign correction
      nr := nr + 512;
    end if;
    return Op & dst & std_logic_vector(to_unsigned(nr, 9));
  end Assemble_RIO;

  function Assemble_RI8  (Op: SLV4; dst: SLV3; val: integer) return SLV16 is
    variable imm : integer := val;
  begin
    assert (Op /= Op_IN) and (Op /= Op_OUT)
        report "Assemble_IO : wrong opcode :" & SLV_to_bin(Op)
         severity failure;

    assert (imm < 256) and (imm >= -128)
        report "Assemble_RI8 : Immediate value is out of range :" & integer'image(imm)
         severity failure;
    if imm < 0 then    -- sign correction
      imm := imm + 256;
    end if;
    return Op & dst & '1' & std_logic_vector(to_unsigned(imm, 8));
  end Assemble_RI8;

  function Assemble_RI3  (Op: SLV4; dst: SLV3; val: integer) return SLV16 is
    variable imm : integer := val;
  begin
    assert (Op /= Op_IN) and (Op /= Op_OUT)
        report "Assemble_IO : wrong opcode :" & SLV_to_bin(Op)
         severity failure;

    assert (imm < 4) and (imm >= -2)
        report "Assemble_RI3 : Immediate value is out of range :" & integer'image(imm)
         severity failure;
    if imm < 0 then    -- sign correction
      imm := imm + 4;
    end if;
    return Op & dst & "000001" & std_logic_vector(to_unsigned(imm, 3));
  end Assemble_RI3;

  function Assemble_RI3C (Op: SLV4; dst: SLV3; val: integer; cond: SLV4) return SLV16 is
    variable imm : integer := val;
  begin
    assert (Op /= Op_IN) and (Op /= Op_OUT)
        report "Assemble_IO : wrong opcode :" & SLV_to_bin(Op)
         severity failure;

    assert (imm < 4) and (imm >= -2)
        report "Assemble_RI3 : Immediate value is out of range :" & integer'image(imm)
         severity failure;
    if imm < 0 then    -- sign correction
      imm := imm + 4;
    end if;
    return Op & dst & '0' & cond & '1' & std_logic_vector(to_unsigned(imm, 3));
  end Assemble_RI3C;

  function Assemble_RR  (Op: SLV4; dst: SLV3; src: SLV3) return SLV16 is
  begin
    assert (Op /= Op_IN) and (Op /= Op_OUT)
        report "Assemble_IO : wrong opcode :" & SLV_to_bin(Op)
         severity failure;

    return Op & dst & "000000" & src;
  end Assemble_RR;

  function Assemble_RRC (Op: SLV4; dst: SLV3; src: SLV3; cond: SLV4) return SLV16 is
  begin
    assert (Op /= Op_IN) and (Op /= Op_OUT)
        report "Assemble_IO : wrong opcode :" & SLV_to_bin(Op)
         severity failure;

    return Op & dst & '0' & cond & '0' & src;
  end Assemble_RRC;

  ------------------------------------------------------------------------
  -- the big all-in-one:

  function Assemble(txt, context: string) return SLV16 is
    variable c, d: character;
    variable i, j, k, l, opc, dest, src, cond: integer := 0;
    type int_array_type is array (1 to 4) of integer;
    variable lengths_array : int_array_type := (0, 0, 0, 0);
    type str_array_type is array (1 to 4) of string (1 to 12);
    variable str_array : str_array_type := (
       "            ", "            ", "            ", "            ");
    variable instruction_slv: SLV16 := "UUUUUUUUUUUUUUUU"; -- invalid result

    function Lookup_opcode(txt: string(1 to 4)) return integer is
      variable i: integer :=Opcode_str_array'high;
    begin
      loop
        exit when Opcode_str_array(i)=txt;
        i := i-1;
        exit when i < 0;
      end loop;
      return i;
    end Lookup_opcode;

    function Lookup_register(txt: string(1 to 2)) return integer is
      variable i: integer :=Reg_str_array'high;
    begin
      loop
        exit when Reg_str_array(i)=txt;
        i := i-1;
        exit when i < 0;
      end loop;
      return i;
    end Lookup_register;

    function Lookup_condition(txt: string(1 to 4)) return integer is
      variable i: integer :=Cond_str_array'high;
    begin
      loop
        exit when Cond_str_array(i)=txt;
        i := i-1;
        assert i >= 0
           report context & "|Assembler : unknown condition name"
           severity failure;
      end loop;
      return i;
    end Lookup_condition;

  begin

    ----------------------------------------------------------------------
    -- first : slice the string into suitable substrings:

    main_parse: loop
      parse_space: loop
        i := i + 1;
        exit main_parse when i > txt'length;

        c := txt(i);
        exit when (c /= ' ') and (c /= HT);
      end loop parse_space;

      -- comment:
      exit main_parse when c=';';

      -- collect a new token:
      j := 0;
      k := k+1;
      assert str_array'high >= k
         report context & "|Parse : too many arguments"
         severity failure;

      parse_word: loop
        d := upcase_lookup(c);
        exit parse_word when d=NUL;

        j := j+1;
        assert j <= 12
           report context & "|Parse : word too long"
           severity failure;
        str_array(k)(j) := d;
        lengths_array(k) := j;

        i := i+1;
        exit main_parse when i > txt'length;
        c := txt(i);
      end loop parse_word;

      assert (c=' ') or (c=HT)
        report context & "|Parse : invalid character '" & c & "'"
        severity failure;
    end loop;

--    report context & "  " & integer'image(k) & "tokens: "
--       & " " & integer'image(lengths_array(1)) & ":" & str_array(1)
--       & " " & integer'image(lengths_array(2)) & ":" & str_array(2)
--       & " " & integer'image(lengths_array(3)) & ":" & str_array(3)
--       & " " & integer'image(lengths_array(4)) & ":" & str_array(4);

    ----------------------------------------------------------------------
    -- Now try to make sens of all of this:

    case k is
      when 0 =>   -- empty line, return nothing
                return "UUUUUUUUUUUUUUUU";

      when 1 =>   -- meta-code ?
        if str_array(1)="NOP         " then
          return "0000000000000000";
        end if;
        if str_array(1)="INV         " then
          return "1111111111111111";
        end if;
        assert false 
           report context & "|Assembler : missing argument ?"
           severity failure;

      when 3|4 => -- instruction ?
        -- 1) opcode
        assert lengths_array(1) <= 4
           report context & "|Assembler : invalid opcode"
           severity failure;
        opc := Lookup_opcode(str_array(1)(1 to 4));
        assert opc >= 0
           report context & "|Assembler : unknown opcode"
           severity failure;
        instruction_slv(15 downto 12) := std_logic_vector(to_unsigned(opc, 4));

        -- 2) Destination : must be 2 char long
        assert lengths_array(2) = 2
           report context & "|Assembler : invalid destination register"
           severity failure;
        dest := Lookup_register(str_array(2)(1 to 2));
        assert dest >= 0
           report context & "|Assembler : unknown register name"
           severity failure;
        instruction_slv(11 downto 9) := std_logic_vector(to_unsigned(dest, 3));

        -- check special case of IN/OUT
        if (opc/2=6) then
          assert k = 3
             report context & "|Assembler : too many operands for IN/OUT"
             severity failure;

          -- check immediate value in 512 bits
          src := Num_to_int(str_array(3)(1 to lengths_array(3)), context & "|Assembler|SRC: ");
          assert (src >= -256) and (src < 512)
             report context & "|Assembler : Imm9 value too large"
             severity failure;
          if src < 0 then
            src := src + 512;
          end if;
          instruction_slv(8 downto 0) := std_logic_vector(to_unsigned(src, 9));

        else

          -- 3) source
          src := -1;
          if lengths_array(3) = 2 then
            src := Lookup_register(str_array(3)(1 to 2));
          end if;
          if src >= 0 then
            instruction_slv(8 downto 0) := "000000" & std_logic_vector(to_unsigned(src, 3));
--          report "selected Reg";

          else
            -- not a register ? try an immediate...
            src := Num_to_int(str_array(3)(1 to lengths_array(3)), context & "|Assembler|SRC: ");
            -- select long/short
            if (src < 4) and (src >= -4) then
              if src < 0 then
                src := src + 8;
              end if;
              instruction_slv(8 downto 0) := "000001" & std_logic_vector(to_unsigned(src, 3));

--            report "selected Imm3";

            else

              assert (src >= -128) and (src < 256)
                 report context & "|Assembler : Imm8 value too large"
                 severity failure;
              if src < 0 then
                src := src + 512;
              end if;
              instruction_slv(8 downto 0) := '1' & std_logic_vector(to_unsigned(src, 8));

--              report "selected Imm8";

            end if; -- immediate

          end if;  -- not a reg

          -- 4) optional condition (added over the 0s of the instruction word)
          if k = 4 then
            assert instruction_slv(8)='0'
              report context & "|Assembler : condition can not be used with long immediate source"
              severity failure;
            l := lengths_array(4);
            assert l=3 or l=4
              report context & "|Assembler : invalid condition"
              severity failure;
            cond := Lookup_condition(str_array(4)(1 to 4));
            instruction_slv(7 downto 4) := std_logic_vector(to_unsigned(cond, 4));
          end if;
        end if; -- if (opc/2=6)

      when others => assert false
        report context & "|Assembler : missing argument ?"
        severity failure;
    end case;

--    report SLV_to_bin(instruction_slv);

    return instruction_slv;
  end Assemble;

end ygrec8_asm;
