-- A3P_DualPortRAM.vhd
-- version 20140510
-- should work for simulation, Altera, Xilinx and Actel/microsemi.
-- A special, tuned version for Actel's A3P is also provided,
-- which shaves many dumb gates from the inferred circuit.
-- PA3's RAM4K9 serves as guidance for the other implementations.
-- * I've added a reset to comply with the tools but it's not necessary.
-- * RAM4K9 supports write-read-through (write data and it appears
--    at the outside as well) but synthesis adds more gates...
--     See WMODEA & WMODEB (high)
--        ### write-read-through is implemented here ###
-- * changed SLV to integer conversion using more standards-compliant code
-- * adapted Write Enable polarity (write='0', read='1')
-- * added some resets
-- * BLKA & BLKB don't seem to be supported by other makers
--     so it's dropped (stuck active=low)
-- * PIPEA & PIPEB tied low (not pipelined)
-- Use the A3P_Slices version for microsemi because it fails
--   to infer correctly and creates tons of FF ! I don't see
--    how to make it accept "write through" which is necessary
--    for miniYASEP.
--
-- version 20171114 : stripped-down for standalone use.
--  See also InferedDualPortRAM.vhdl
--
-- from PA3_HB.pdf v1.5, p6-8:
-- WMODEA and WMODEB
-- These signals are used to configure the behavior of the output
-- when the RAM is in write mode. A LOW on these signals makes
-- the output retain data from the previous read. A HIGH indicates
-- pass-through behavior, wherein the data being written will appear
-- immediately on the output. This signal is overridden when the RAM is being read.

library ieee;
    use ieee.std_logic_1164.all;
    use ieee.numeric_std.all;
library proasic3;
    use proasic3.all;

entity bram is
  generic (
    SLICES  : integer := 4;
    SLICE_WIDTH : integer := 4; 
    WIDTH   : integer := 16; -- please preserve SLICES*SLICE_WIDTH;
    ADRBITS : integer := 10
    -- For Microsemi :
    -- ADRBITS       9  10  11  12 
    -- SLICE_WIDTH  8/9  4   2   1
  );
  port (
    reset : in std_logic := '1'; -- may be omitted
    -- port A :
    a_clk, a_wr_en : in std_logic;
    a_addr : in std_logic_vector(ADRBITS-1 downto 0);
    a_din  : in std_logic_vector(WIDTH-1 downto 0);
    a_dout : out std_logic_vector(WIDTH-1 downto 0);
    -- port B :
    b_clk, b_wr_en : in std_logic;
    b_addr : in std_logic_vector(ADRBITS-1 downto 0);
    b_din  : in std_logic_vector(WIDTH-1 downto 0);
    b_dout : out std_logic_vector(WIDTH-1 downto 0)
  );
end bram;

  
architecture A3P_Slices of bram is
  signal adra, adrb: std_logic_vector(11 downto 0) := "000000000000";
  signal W1, W0 : std_logic;

  component RAM4K9
    port(ADDRA11, ADDRA10, ADDRA9, ADDRA8, ADDRA7, ADDRA6, 
         ADDRA5,  ADDRA4,  ADDRA3, ADDRA2, ADDRA1, ADDRA0,
         ADDRB11, ADDRB10, ADDRB9, ADDRB8, ADDRB7, ADDRB6,
         ADDRB5,  ADDRB4,  ADDRB3, ADDRB2, ADDRB1, ADDRB0,
         DINA8, DINA7, DINA6, DINA5, DINA4, DINA3, DINA2, DINA1, DINA0,
         DINB8, DINB7, DINB6, DINB5, DINB4, DINB3, DINB2, DINB1, DINB0,
         WIDTHA0, WIDTHA1, WIDTHB0, WIDTHB1, PIPEA, PIPEB, WMODEA, WMODEB, 
         BLKA, BLKB, WENA, WENB, CLKA, CLKB, RESET : in std_logic := 'U';
         DOUTA8, DOUTA7, DOUTA6, DOUTA5, DOUTA4, DOUTA3, DOUTA2, DOUTA1, DOUTA0,
         DOUTB8, DOUTB7, DOUTB6, DOUTB5, DOUTB4, DOUTB3, DOUTB2, DOUTB1, DOUTB0 : out std_logic) ;
    end component;

begin
  adra(a_addr'range) <= a_addr;
  adrb(b_addr'range) <= b_addr;
  adra(11 downto ADRBITS) <= (others=>'0');
  adrb(11 downto ADRBITS) <= (others=>'0');

  W0 <= '0' when SLICE_WIDTH=1 or SLICE_WIDTH=4
   else '1' when SLICE_WIDTH=2 or SLICE_WIDTH=8 or SLICE_WIDTH=9
   else 'X'; -- error !
  W1 <= '0' when SLICE_WIDTH=1 or SLICE_WIDTH=2
   else '1' when SLICE_WIDTH=4 or SLICE_WIDTH=8 or SLICE_WIDTH=9 
   else 'X'; -- error !

  iterate_slices:
  for i in SLICES-1 downto 0 generate
    signal tmpa, tmpb, outa, outb: std_logic_vector( 8 downto 0);
  begin
    -- adapt the widths with a vector where some bits may be dropped
    -- (this simplifies the instantiation code below)
    tmpa(SLICE_WIDTH-1 downto 0) <= a_din (((i+1)*SLICE_WIDTH)-1 downto i*SLICE_WIDTH);
    tmpb(SLICE_WIDTH-1 downto 0) <= b_din (((i+1)*SLICE_WIDTH)-1 downto i*SLICE_WIDTH);
    tmpa(8 downto SLICE_WIDTH) <= (others=>'0');
    tmpb(8 downto SLICE_WIDTH) <= (others=>'0');
    a_dout(((i+1)*SLICE_WIDTH)-1 downto i*SLICE_WIDTH) <= outa(SLICE_WIDTH-1 downto 0);
    b_dout(((i+1)*SLICE_WIDTH)-1 downto i*SLICE_WIDTH) <= outb(SLICE_WIDTH-1 downto 0);

    DPRAMslice: RAM4K9
    port map(
      ADDRA11=>adra(11),
      ADDRA10=>adra(10),
      ADDRA9=>adra(9),
      ADDRA8=>adra(8),
      ADDRA7=>adra(7),
      ADDRA6=>adra(6),
      ADDRA5=>adra(5),
      ADDRA4=>adra(4),
      ADDRA3=>adra(3),
      ADDRA2=>adra(2),
      ADDRA1=>adra(1),
      ADDRA0=>adra(0),
      ADDRB11=>adrb(11),
      ADDRB10=>adrb(10),
      ADDRB9=>adrb(9),
      ADDRB8=>adrb(8),
      ADDRB7=>adrb(7),
      ADDRB6=>adrb(6),
      ADDRB5=>adrb(5),
      ADDRB4=>adrb(4),
      ADDRB3=>adrb(3),
      ADDRB2=>adrb(2),
      ADDRB1=>adrb(1),
      ADDRB0=>adrb(0),
      DINA8=>tmpa(8),
      DINA7=>tmpa(7),
      DINA6=>tmpa(6),
      DINA5=>tmpa(5),
      DINA4=>tmpa(4),
      DINA3=>tmpa(3),
      DINA2=>tmpa(2),
      DINA1=>tmpa(1),
      DINA0=>tmpa(0),
      DINB8=>tmpb(8),
      DINB7=>tmpb(7),
      DINB6=>tmpb(6),
      DINB5=>tmpb(5),
      DINB4=>tmpb(4),
      DINB3=>tmpb(3),
      DINB2=>tmpb(2),
      DINB1=>tmpb(1),
      DINB0=>tmpb(0),
      WIDTHA0 => W0,
      WIDTHA1 => W1,
      WIDTHB0 => W0,
      WIDTHB1 => W1, 
      PIPEA=>'0',
      PIPEB=>'0',
      WMODEA=>'1',
      WMODEB=>'1',
      BLKA=>'0',
      BLKB=>'0',
      WENA=>a_wr_en,
      WENB=>b_wr_en,
      CLKA=>a_clk,
      CLKB=>b_clk,
      RESET=>reset,
      DOUTA8=>outa(8),
      DOUTA7=>outa(7),
      DOUTA6=>outa(6),
      DOUTA5=>outa(5),
      DOUTA4=>outa(4),
      DOUTA3=>outa(3),
      DOUTA2=>outa(2),
      DOUTA1=>outa(1),
      DOUTA0=>outa(0),
      DOUTB8=>outb(8),
      DOUTB7=>outb(7),
      DOUTB6=>outb(6),
      DOUTB5=>outb(5),
      DOUTB4=>outb(4),
      DOUTB3=>outb(3),
      DOUTB2=>outb(2),
      DOUTB1=>outb(1),
      DOUTB0=>outb(0)
    );
  end generate;
end A3P_Slices;
