-- YGREC8/Core.vhdl
-- created dim. nov. 12 22:26:45 CET 2017 by Yann Guidon (whygee@f-cpu.org)
-- version mar. nov. 14 03:10:08 CET 2017
-- version mar. nov. 14 04:34:04 CET 2017 : yeah ! datapath written !
-- version mer. nov. 15 07:42:50 CET 2017 : first test, first instruction : first flaw found.
-- version jeu. nov. 16 00:11:47 CET 2017 : flags writeback flaw corrected (CMPU/CMPS works better)
-- version jeu. nov. 16 03:22:14 CET 2017 : ADR_PORT removed,  uses SLV shorthands from YGREC8/YGREC8_def.vhdl
-- version ven. nov. 17 02:47:03 CET 2017 : polishing the conditions
-- version mar. déc.  5 21:36:40 CET 2017 : better synthesis, tigher definitions, fewer dangling signals
-- version mar. jan.  2 03:46:35 CET 2018 : CARRY IN modified for Imm3 range extension
-- version dim. jan. 28 03:13:47 CET 2018 : LDCx, OVL, ...
--
-- Released under the GNU AGPLv3 license
--
-- This is the main file of YGREC8,
-- it assembles the sub-units to make the core's datapath.
-- Many signals and registers are set as inout mode to ease coding.

Library ieee;
    use ieee.std_logic_1164.all;
Library work;
    use work.all;
    use work.ygrec8_def.all;

entity YGREC8_Core is
  port(
    clk, reset, enable_cycle, CB0, CB1, CB2, CB3 : in SL;
    InstructionWord : in SLV16;
    IN_PORT : in SLV8;
    -- actually, these are registers:
    PC, NextPC, WriteBack, DST, SRC : inout SLV8;
    nC, C, nS, S, nZ, Z, write_enable, OVL : inout SL);
end YGREC8_Core;

architecture A3P of YGREC8_Core is
  signal AdrSRC, AdrWr, AdrDly, AdrDST, Condition : SLV3;
  signal Opcode, CBl : SLV4;
  signal Imm8, Imm3ext, PC1, SRC_reg, SHL, SUM, ROP2, Result,
      D1, D2, A1, A2, CondVect : SLV8;
  signal RI3, RI8, Cin, Cout, disable_write, flag_enable,
    a_wr_en, b_wr_en, CondNeg, CondSel,
    LDCstate, RI8dly : SL;
  signal a_addr, b_addr : SLV9;
begin
  -- alias the instruction fields to the instruction word :
  Imm8      <= InstructionWord(15 downto  8) when RI8dly='1'
          else InstructionWord( 7 downto  0);
  AdrSRC    <= InstructionWord( 2 downto  0);
  Imm3ext   <= (    0=>InstructionWord(0),
                    1=>InstructionWord(1),
               others=>InstructionWord(2));
  RI3       <= InstructionWord( 3);
  Condition <= InstructionWord( 6 downto  4);
  CondNeg   <= InstructionWord( 7);
  RI8       <= '1' when LDCstate='1'
          else InstructionWord( 8);
  AdrDST    <= InstructionWord(11 downto  9);
  AdrWr     <= AdrDly when LDCstate='1'
          else AdrDST;
  Opcode    <= Op_MOV when LDCstate='1'
          else InstructionWord(15 downto 12);

  -- the output OVL signal is looped back to enable_cycle through an external FSM.
  OVL <= '1' when Opcode=Op_CALL
         and AdrDST="111"
     else '0';

  entity_INC_PC: entity INC8 port map (
     A => PC,
     Y => PC1,
     V => open);  -- => used to trigger a trap later

  WriteBack <= PC when Opcode=Op_CALL
          else Result;
  NextPC <= Result when (Opcode=Op_LDCL) -- but not registered
                or (((Opcode=Op_CALL) or (AdrDST=Reg_PC)) and disable_write='0')
    else PC1;

  entity_REG8: entity REG8 port map (
    clk => clk,
    reset => reset,
    write_enable => write_enable,
    AdrSRC => AdrSRC,
    AdrDST => AdrDST,
    AdrWr  => AdrWr,
    PC => PC,
    D1 => D1,
    D2 => D2,
    RESULT => WriteBack,
    SRC => SRC_reg,
    DST => DST,
    A1 => A1,
    A2 => A2
  );

--  OUT_PORT <= DST;
  SRC <= SRC_reg when (RI8='0' and RI3='0')  -- slow path first
    else Imm8  when RI8='1'                  -- fast lane next
    else Imm3ext; -- (RI8='0' and RI3='1')


**************************************************************************
  Cin <= '0' when Opcode(1 downto 0)=Op_ADD(1 downto 0)
        and ( RI8='1' or (RI3='1' and InstructionWord(2)='1' ))   -- don't INC if ADD Imm8, or Imm3 negative
    else '1';  -- no ADDC or SUBB support. yet ?

  entity_ALU8: entity ALU8 port map (
     F0 => Opcode(0),
     F1 => Opcode(1),
     F2 => Opcode(2),
--     F3 => Opcode(3),
     Cin => Cin,
     SRC => SRC,
     DST => DST,
     SUM => SUM,
     ROP2 => ROP2,
     Cout => Cout);

  entity_SHL8: entity SHL8 port map (
     F0 => Opcode(0),
     F1 => Opcode(1),
     SRC => SRC(2 downto 0),
     DST => DST,
     SHL => SHL);

  Result <= SUM  when Opcode(3 downto 2)=Op_ADD(3 downto 2) -- ADD, SUB, CMPU, CMPS
       else SHL  when Opcode(3 downto 2)=Op_SHL(3 downto 2) -- SHR, SHL, SAR,  ROL
       else IN_PORT when Opcode=Op_IN                       -- IN (OUT does not write)
       else ROP2 when Opcode(3 downto 2)=Op_OR (3 downto 2) -- OR, XOR, AND, ANDN
       else SRC;                                            -- MOV, CALL

  nC <= Cout;
  nS <= Result(7);
  nZ <= '0' when Result="00000000"
   else '1';

  disable_write <= '1' when
            (Opcode=Op_IN and RI8='1') -- Op_OUT
          OR Opcode=Op_CMPU
          OR Opcode=Op_CMPS
          OR Opcode=Op_LDCL  -- write during the next cycle with LDCstate='1' which forces Opcode=Op_MOV
          OR enable_cycle='0'
    else '0';

  -- This is now rewritten as a vector/balanced MUX
--  CondSel <= '1' when Condition=Cond_A(2 downto 0)
--        else  C  when Condition=Cond_C(2 downto 0)
--        else  S  when Condition=Cond_S(2 downto 0)
--        else  Z  when Condition=Cond_Z(2 downto 0)
--        else CB0 when Condition=Cond_0(2 downto 0)
--        else CB1 when Condition=Cond_1(2 downto 0)
--        else CB2 when Condition=Cond_2(2 downto 0)
--        else CB3 ; -- Condition=Cond_3(2 downto 0)
  CondVect <= CBl & Z & S & C & '1';

  write_enable <= (CondSel xor CondNeg) or RI8 
           when disable_write='0' else '0';
  -- special case for the flags and CMPU/CMPS:
  flag_enable <= ((CondSel xor CondNeg) or RI8)
          when (Opcode/=Op_OUT AND enable_cycle='1')
       else '0';

  -- The register set has its own registers,
  -- this process only has to deal with the flags and PC.
  process (clk, reset)
  begin
    if reset='0' then
       C <= '0';
       S <= '0';
       Z <= '0';
      CBl <= "0000";
      PC <= "00000000";
      LDCstate <= '0';
      RI8dly   <= '0';
      AdrSRCdly <= "000";
    else
      if rising_edge(clk) then
        if enable_cycle='1' then
          CBl <= CB3 & CB2 & CB1 & CB0; -- prevent glitches
          AdrDly <= AdrWr;
          if (Opcode=Op_LDCL) then
            LDCstate <= '1';
            RI8dly   <= InstructionWord( 8); -- RI8;
          else
            LDCstate <= '0';
            RI8dly   <= '0';
            PC <= NextPC;

            -- a modified version of write_enable
            -- that allows CMPU/CMPS to update the flags:
            if flag_enable='1' then
              -- update only if DST register is written
              S <= nS;
              Z <= nZ;
              if Opcode(3 downto 2)=Op_ADD(3 downto 2) then
                -- Update Carry only if ADD, SUB, CMPU, CMPS
                C <= nC;
              end if;
            end if;
          end if;
        end if;
      end if;
    end if;
  end process;

  -- Memory system:

  -- The MSB should be configurable somehow under program control...
  a_addr(8) <= '0';
  b_addr(8) <= '1';

  a_addr(7 downto 0) <= WriteBack when AdrDST=Reg_A1 and write_enable='1'
       else A1;
  b_addr(7 downto 0) <= WriteBack when AdrDST=Reg_A2 and write_enable='1'
       else A2;

  a_wr_en <= '0' when AdrDST=Reg_D1 and write_enable='1'
        else '1';
  b_wr_en <= '0' when AdrDST=Reg_D2 and write_enable='1'
        else '1';

  entity_bram: entity bram 
  generic map(
    SLICES      => 1,
    SLICE_WIDTH => 8,
    WIDTH       => 8, -- please preserve SLICES*SLICE_WIDTH;
    ADRBITS     => 9
    -- For Microsemi :
    -- ADRBITS       9  10  11  12
    -- SLICE_WIDTH  8/9  4   2   1
  )
  port map (
    reset   => reset,
    -- port A :
    a_clk   => clk,
    a_wr_en => a_wr_en,
    a_addr  => a_addr,
    a_din   => WriteBack,
    a_dout  => D1,
    -- port B :
    b_clk   => clk,
    b_wr_en => b_wr_en,
    b_addr  => b_addr,
    b_din   => WriteBack,
    b_dout  => D2 );

end A3P;
